{ UZ80 - Z80 emulator                                            }
{ =-=-=-=-=-=-=-=-=-=-                                           }
{ by pscardoso@lme.usp.br                                        }
{ do not change the source, this file is automatically generated }
{ if you find a bug, send information to the author.             }

{ Last changes }
{ 30/MAY/1998 - Fixed LD A, R instruction. Now spectrum snapshot  
                Cobra is working.                                }
{ 29/MAY/1998 - Fixed the IN R,(C) instruction  (the 16 bit port  
                address needs to be BC). Now ZXSpectrum keyboard  
                is working.                                      }

{ 21/MAY/1998 - Public release (version 0.21)                     }
{ 21/MAY/1998 - Fixed R register incremention. Now ladybug works. }
{ 14/MAY/1998 - Fixed IN R,(X) (this instruction affect flags!    }
{ 07/MAY/1998 - Fixed LDIR, CPIR, OTIR, INIR, etc (now Galaxians  
                bee attack us again.                              }

{ 06/MAY/1998 - Public release (version 0.2)                      }
{ 04/MAY/1998 - implementation of INI, IND, INIR, INDR            }
{ 03/MAY/1998 - implementation of CPI, CPD, CPIR, CPDR            }
{ 02/MAY/1998 - implementation of OUTI, OTIR, OUTD, OTDR          }
{ 30/APR/1998 - fixed bugs in CCF                                 }
{ 28/APR/1998 - fixed LDIR, LDDR functions (problems with delphi) }
{ 27/APR/1998 - now EvalXXXInstr preservs (e)si, (e)di, (e)bx.    }
{ 16/APR/1998 - First public release (version 0.1)                }

unit UZ80;

interface

const
  IGNORE_INT = $FF;
  IPERIOD    = $50000;
  S_FLAG = $80;
  Z_FLAG = $40;
  U2FLAG = $20;
  H_FLAG = $10;
  U1FLAG = $08;
  P_FLAG = $04;
  V_FLAG = $04;
  N_FLAG = $02;
  C_FLAG = $01;
  MAX_PROCS     = 10;
  FREG          = 00;
  AREG          = 01;
  CREG          = 02;
  BREG          = 03;
  EREG          = 04;
  DREG          = 05;
  LREG          = 06;
  HREG          = 07;
  IXLREG        = 08;
  IXHREG        = 09;
  IYLREG        = 10;
  IYHREG        = 11;
  PCLREG        = 12;
  PCHREG        = 13;
  SPLREG        = 14;
  SPHREG        = 15;
  F1LREG        = 16;
  A1HREG        = 17;
  C1LREG        = 18;
  B1HREG        = 19;
  E1LREG        = 20;
  D1HREG        = 21;
  L1LREG        = 22;
  H1HREG        = 23;
  IFF1          = 24;
  IFF2          = 25;
  IREG          = 26;
  RREG          = 27;
  MAXROML       = 28;
  MAXROMH       = 29;
  MAXRAML       = 30;
  MAXRAMH       = 31;
  PREVPCH       = 32;
  PREVPCL       = 33;
  TICKCOUNTH    = 34;
  TICKCOUNTL    = 35;
  TICKCOUNT32H  = 36;
  TICKCOUNT32L  = 37;
  TICKMACHH     = 38;
  TICKMACHL     = 39;
  TICKMACH32H   = 40;
  TICKMACH32L   = 41;
  MEMORY        = 42;
  GETBYTEFUNC   = 46;
  PUTBYTEFUNC   = 50;
  OUTBYTEFUNC   = 54;
  INBYTEFUNC    = 58;
  INTERRUPTFUNC = 62;
  MAXRAMXXX     = 66;
  NMIINT        = 68;
  HALTED        = 70;
  IM            = 72;
  PENDINGNMI    = 74;
  PENDINGIRQ    = 78;
  AFREG         = 00;
  BCREG         = 02;
  DEREG         = 04;
  HLREG         = 06;
  IXREG         = 08;
  IYREG         = 10;
  PCREG         = 12;
  SPREG         = 14;
  AF1REG        = 16;
  BC1REG        = 18;
  DE1REG        = 20;
  HL1REG        = 22;
  IFFREG        = 24;
  IRREG         = 26;
  MAXROM        = 28;
  MAXRAM        = 30;
  PREVPC        = 32;
  TICKCOUNT     = 34;
  TICKCOUNT32   = 36;
  TICKMACH      = 38;
  TICKMACH32    = 40;

type
  TPair = record
    case byte of
      0: (l, h: byte);
      1: (w: Word);
  end;
  PZ80ASMRegs = ^TZ80ASMRegs;
  TZ80ASMRegs = record
    AF,BC,DE,HL,IX,IY,PC,SP: TPair;
    AF1,BC1,DE1,HL1: TPair;
    IFF1, IFF2: Byte;
    I, R: Byte;
    nMaxRom, nMaxRam, nPrevPC: Word;
    nTicks: LongInt;
    nTickMach: LongInt;
  end;
  TProcessor = record
    aRegs: array[0..MEMORY-1] Of Byte;
    oMemory: Pointer;
    oGetByteFunc: function (nEnder: Word): Byte;
    oPutByteFunc: procedure (nEnder, nValue: Word);
    oOutByteFunc: procedure (nEnder, nValue: Word);
    oInByteFunc: function (nEnder: Word): Byte;
    oInterruptFunc: procedure;
    nMaxRAM, nNMIInt, nHalted, nIM: Word;
    nPendingNMI, nPendingIRQ: LongInt;
  end;

var
  aProcs: array[1..MAX_PROCS] of TProcessor;

procedure EvalInstr(nProc: Word);
procedure EvalOneInstr(nProc: Word);

implementation

const
 nStructSize = SizeOf(TProcessor);
 {$I TABLES.INC}

procedure OpCode; forward;
procedure ExecOpCode; forward;
procedure OpCodeDDCB; forward;
procedure OpCodeFDCB; forward;
procedure OpCodeCB; forward;
procedure OpCodeDD; forward;
procedure OpCodeED; forward;
procedure OpCodeDDED; forward;
procedure OpCodeFDED; forward;
procedure OpCodeFD; forward;

procedure PutByte; assembler;
asm
  cmp   bx, word ptr si[MAXROM]
  jb    @Exit
  pusha
  xor   ah, ah
  push  bx
  push  ax
  call  dword ptr si[PUTBYTEFUNC]
  popa
@Exit:
end;

procedure OutByte; assembler;
asm
  pusha
  xor   ah, ah
  push  bx
  push  ax
  call  dword ptr si[OUTBYTEFUNC]
  popa
end;

procedure GetByte; assembler;
asm
  cmp   word ptr si[MAXRAM], $FFFF
  je    @UseFunc
  cmp   bx, word ptr si[MAXRAM]
  ja    @UseFunc
  push  es
  push  di
  mov   di, word ptr si[MEMORY+0]
  mov   es, word ptr si[MEMORY+2]
  mov   al, byte ptr es:[di+bx]
  pop   di
  pop   es
  jmp   @Exit
@UseFunc:
  push  bx
  push  cx
  push  dx
  push  si
  push  di
  push  es
  push  ds

  mov   dx, ax
  push  dx
  push  bx
  call  dword ptr si[GETBYTEFUNC]
  pop   dx
  mov   ah, dh

  pop   ds
  pop   es
  pop   di
  pop   si
  pop   dx
  pop   cx
  pop   bx
@Exit:
end;

procedure InByte; assembler;
asm
  push  bx
  push  cx
  push  dx
  push  si
  push  di
  push  es
  push  ds

  mov   dx, ax
  push  dx
  push  ax
  call  dword ptr si[INBYTEFUNC]
  pop   dx
  mov   ah, dh

  pop   ds
  pop   es
  pop   di
  pop   si
  pop   dx
  pop   cx
  pop   bx
end;

procedure EIInterrupt; assembler;
asm
  pusha
  cmp  byte ptr si[PENDINGIRQ+3], $01
  jne  @DoInterrupt

  cmp  word ptr si[PENDINGNMI], 0
  je   @Exit

@DoInterrupt:

  cmp  word ptr si[PENDINGNMI], 0
  jne  @StartInt

  cmp  word ptr si[IFF1], 0
  je   @Exit

@StartInt:
  mov  word ptr si[PREVPC], $FFFF
  mov  word ptr si[IFF1], 0

  cmp  word ptr si[HALTED], 1  
  jne  @IsNotHalted

  inc  word ptr si[PCREG]
  mov  word ptr si[HALTED], 0

@IsNotHalted:

  cmp  word ptr si[PENDINGNMI], 0
  je   @NoPendingNMI
  { PUSH PC }
  mov  word ptr si[PENDINGNMI], 0
  dec  word ptr si[SPREG]
  mov  bx, word ptr si[SPREG]
  mov  al, byte ptr si[PCHREG]
  call PutByte
  dec  word ptr si[SPREG]
  mov  bx, word ptr si[SPREG]
  mov  al, byte ptr si[PCLREG]
  call PutByte
  mov  ax, word ptr si[NMIINT]
  mov  word ptr si[PCREG], ax
  jmp  @Exit

@NoPendingNMI:
  mov  byte ptr si[PENDINGIRQ+3], $01

  cmp  byte ptr si[IM], 0
  je   @IsIM0                   { Interruption mode is 0 }

  cmp  byte ptr si[IM], 3
  je   @IsIM2                   { Interruption mode is 2 }

  { IM-undoc, IM1 }
  { push PC }
  { PC <- 38h }
  dec  word ptr si[SPREG]
  mov  bx, word ptr si[SPREG]
  mov  al, byte ptr si[PCHREG]
  call PutByte
  dec  word ptr si[SPREG]
  mov  bx, word ptr si[SPREG]
  mov  al, byte ptr si[PCLREG]
  call PutByte
  mov  word ptr si[PCREG], $0038
  je   @Exit

@IsIM0:
  cmp  byte ptr si[PENDINGIRQ+2], $CD
  jne  @IsNotCall
  { push PC }
  { PC <- 38h }
  dec  word ptr si[SPREG]
  mov  bx, word ptr si[SPREG]
  mov  al, byte ptr si[PCHREG]
  call PutByte
  dec  word ptr si[SPREG]
  mov  bx, word ptr si[SPREG]
  mov  al, byte ptr si[PCLREG]
  call PutByte
  mov  word ptr si[PCREG], $FFFF
  je   @Exit

@IsNotCall:
  cmp  byte ptr si[PENDINGIRQ+2], $C3
  jne  @IsNotJump
  mov  word ptr si[PCREG], $FFFF
  je   @Exit

@IsNotJump:
  xor  ax, ax
  mov  al, byte ptr si[PENDINGIRQ+0]
  call ExecOpCode

  jmp  @Exit

@IsIM2:
  { IM2 }
  { push PC }
  { PC <- Mem[I shl 8 + FF] }
  dec  word ptr si[SPREG]
  mov  bx, word ptr si[SPREG]
  mov  al, byte ptr si[PCHREG]
  call PutByte
  dec  word ptr si[SPREG]
  mov  bx, word ptr si[SPREG]
  mov  al, byte ptr si[PCLREG]
  call PutByte
  mov  bh, byte ptr si[IREG]
  mov  bl, byte ptr si[PENDINGIRQ]
  call GetByte
  xchg ah, al
  inc  bx
  call GetByte
  xchg ah, al
  mov  word ptr si[PCREG], ax

@Exit:

  popa
end;

procedure Interrupt; assembler;
asm
  pusha
  call dword ptr si[INTERRUPTFUNC]
  call EIInterrupt
  popa
end;


{$I OPCODE.INC}
{$I OPCDDDCB.INC}
{$I OPCDFDCB.INC}
{$I OPCODECB.INC}
{$I OPCODEDD.INC}
{$I OPCODEED.INC}
{$I OPCDDDED.INC}
{$I OPCDFDED.INC}
{$I OPCODEFD.INC}

procedure EvalInstr(nProc: Word);
var
  nStrOffs: Word;
begin
  nStrOffs := nStructSize*(nProc-1);
  asm
    (* cli; *)
    push si
    push di
    push bx
    mov  si, offset aProcs
    add  si, nStrOffs

    mov  word ptr si[TICKCOUNT], 0    { initialize cycles counter }
@LoopGame:
    inc  byte ptr si[RREG]
    call OpCode

    mov  ax, word ptr si[TICKCOUNT]
    cmp  ax, 50000

    jb   @LoopGame
    (* STI *)
    call Interrupt
    pop  bx
    pop  di
    pop  si
  end;
end;

procedure EvalOneInstr(nProc: Word);
var
  nStrOffs: Word;
begin
  nStrOffs := nStructSize*(nProc-1);
  asm
    push si
    push di
    push bx
    mov  si, offset aProcs
    add  si, nStrOffs
    inc  byte ptr si[RREG]
    call OpCode

    mov  ax, word ptr si[TICKCOUNT]
    cmp  ax, 50000
    jb   @NoInt

    call Interrupt

    mov  word ptr si[TICKCOUNT], 0    { initialize cycles counter }
@NoInt:
    pop  bx
    pop  di
    pop  si
  end;
end;

end.
